/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.netbios;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Provides logging methods specific to the netbios package. Log class
 * members are applied to the Log's mask(set as an arithmetic expression
 * from {@link jcifs.util.Log#setMask(int mask)}) to "flip on" logging of
 * netbios specific information.
 * <p><blockquote><pre>
 *     Log.setMask( Log.EXCEPTIONS +
 *                 jcifs.netbios.Log.PACKET_DATA );
 * </pre></blockquote>
 * <p>
 * See the {@link jcifs.util.Log} parent class for details about this
 * logging style.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see       jcifs.util.Log
 * @since     jcifs-0.1
 */

public class Log extends jcifs.util.Log {

    // supress javadoc constructor summary
    Log() {}

/**
 *
 * This mask produces limited netbios name service packet information. See
 * <a href="http://www.cis.ohio-state.edu/rfc/rfc1002.txt">RFC 1002</a>
 * for a detailed description of packet contents and their meaning.
 * <p><blockquote><pre>
 *  Mar 13 09:17:04 - name service packet
 *   angus.mimosa.com/192.168.1.152:137 [62]
 *   28887  Positive Name Query Response
 *   authoritative answer,unicast,recursion desired,recursion available
 * </pre></blockquote>
 *
 * @see jcifs.util.Log#setMask(int mask)
 */

    public static final int PACKET_DATA     = 0x00000010;

/**
 * This mask produces detailed netbios name service packet diagrams like
 * those found in <a href="http://www.cis.ohio-state.edu/rfc/rfc1002.txt">RFC
 * 1002</a>.
 * <p><blockquote><pre>
 * Mar 13 09:17:04 - name service packet diagram
 *                       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3
 *   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |            2                  |0| QUERY |0| |R|0|   |B|  0x0  |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |          0x0001               |           0x0000              |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |          0x0000               |           0x0000              |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                                                               |
 *  /  ANGUS<00>                                                    /
 *  |                                                               |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |          NB (0x0020)          |        IN (0x0001)            |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * </pre></blockquote>
 *
 * @see jcifs.util.Log#setMask(int mask)
 */

    public static final int PACKET_DIAGRAMS = 0x00000020;

/**
 * This mask produces the contents of the address cache.
 *
 * @see jcifs.netbios.NbtConfig#setCachePolicy(int cachePolicy)
 * @see jcifs.util.Log#setMask(int mask)
 */

    public static final int ADDRESS_CACHE   = 0x00000040;

    static void printPacketData( String desc, ClientNameServicePacket nsp ) {
        try {
            if(( PACKET_DATA & mask ) == 0 )
                return;
            StringBuffer sb = new StringBuffer();
            sb.append( ' ' ).append( nsp.getAddress() ).append( ":" ).
                        append( nsp.getPort() ).append( " [" ).
                        append( nsp.getLength() ).append( "]" ).
                        append( NEWLINE ).append( ' ' );
            String s = Integer.toString( nsp.getNameTrnId() );
            sb.append( s );
            for( int i = s.length(); i < 7; i++ )
                sb.append( ' ' );
            switch( nsp.getPacketType() ) {    
                case ClientNameServicePacket.POS_NAME_QUERY_RESPONSE:
                    sb.append( "Positive Name Query Response" );
                    break;
                case ClientNameServicePacket.NEG_NAME_QUERY_RESPONSE:
                    sb.append( "Negative Name Query Response" );
                    break;
                case ClientNameServicePacket.REDIRECT_NAME_QUERY_RESPONSE:
                    sb.append( "Redirect Name Query Response" );
                    break;
                case ClientNameServicePacket.WAIT_FOR_ACK_RESPONSE:
                    sb.append( "Wait For Acknowledgement Response" );
                    break;
                case ClientNameServicePacket.NODE_STATUS_RESPONSE:
                    sb.append( "Node Status Response" );
                    break;
                case ClientNameServicePacket.NAME_QUERY_REQUEST:
                    sb.append( "Name Query Request" );
                    break;
                case ClientNameServicePacket.NODE_STATUS_REQUEST:
                    sb.append( "Node Status Request" );
                    break;
                case ClientNameServicePacket.OTHER:
                    sb.append( "Other" );
            }
            sb.append( NEWLINE ).append( ' ' );
            String a[] = new String[5];
            int i = 0;
            if( nsp.isRecurAvailable() )
                a[i++] = new String( "recursion available" );
            if( nsp.isRecurDesired() )
                a[i++] = new String( "recursion desired" );
            if( nsp.isBroadcast() ) {
                a[i++] = new String( "broadcast" );
            } else {
                a[i++] = new String( "unicast" );
            }
            if( nsp.isAuthAnswer() )
                a[i++] = new String( "authoritative answer" );
            while( i > 1 )
                sb.append( a[--i] ).append( ',' );
            sb.append( a[0] );
            switch( nsp.getResultCode() ) {
                case 1:
                    sb.append( NEWLINE ).append( "format error" );
                case 2:
                    sb.append( NEWLINE ).append( "format error" );
                case 4:
                    sb.append( NEWLINE ).append( "unsupported request" );
                case 5:
                    sb.append( NEWLINE ).append( "refused" );
                case 6:
                    sb.append( NEWLINE ).append( "active error" );
                case 7:
                    sb.append( NEWLINE ).append( "conflict error" );
            }
            out.println( desc, sb.toString() );
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.Log.printPacketData()", e );
        }
    }
    static void printAddressCache( String desc, Hashtable addressCache ) {
        try {
            if(( ADDRESS_CACHE & mask ) == 0 )
                return;
            StringBuffer sb = new StringBuffer();
            NbtAddress.CacheEntry ce;
            for( Enumeration e = addressCache.elements(); e.hasMoreElements(); ) {
                ce = (NbtAddress.CacheEntry)e.nextElement();
                sb.append( ' ' ).append( ce.address ).append( '\n' );
            }
            out.println( desc, sb.toString() );
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.Log.printAddressCache()", e );
        }
    }

    private static final char[] COMPOSITE = new String(
        "                      1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3\n" +
        "  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |0| QUERY |0| |1|0|   |0|  0x0  |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |          0x0000               |           0x0000              |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |          0x0000               |           0x0000              |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                                                               |\n" +
        " /                                                               /\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               +\n" +
        " |                                                               |\n" +
        " /   Packet diagram not fully implemented. If you can obtain a   /\n" +
        " /        hex dump, please send it to mballen@erols.com          /\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();

    private static final char[] POS_NAME_QUERY_RESPONSE_TAIL = new String(
        " |                               |U| H |                         |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] WACK_TAIL = new String(
        " |                               |0| QUERY |0| |1|0|   |0|  0x0  |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] NODE_STATUS_RESPONSE_1 = new String(
        " |                               |               |               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+               +\n" +
        " |                                                               |\n" +
        " +                                                               +\n" ).toCharArray();
    private static final char[] NODE_STATUS_RESPONSE_2 = new String(
        " +                                                               +\n" +
        " |                                                               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " | MAC Address = 00-00-00-00-00-00                               |\n" +
        " + 00 bytes follow               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" +
        " | 00 bytes total of statistics  |           0x0000              |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] VARIABLE_STATS_1 = new String(
        " |          0x0000               |           0x0000              |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] VARIABLE_STATS_2 = new String(
        " |          0x0000               |\n" +
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();

    private static final char[] SPACE = new String(
        "                                                                " ).toCharArray();
    private static final char[] PLUS_LINE = new String(
        " +                                                               +\n" ).toCharArray();
    private static final char[] SLASH_LINE = new String(
        " /                                                               /\n" ).toCharArray();
    private static final char[] BAR_LINE = new String(
        " |                                                               |\n" ).toCharArray();
    private static final char[] FULL_LINE = new String(
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] HALF_LINE = new String(
        " +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n" ).toCharArray();
    private static final char[] QUARTER_LINE = new String(
        " +-+-+-+-+-+-+-+-+\n" ).toCharArray();

    //NAME_QUERY_REQUST
    static char[] getNameQueryRequestTemplate() {
        char dst[] = new char[1003];
        System.arraycopy( COMPOSITE, 0, dst, 0, 1003 );
        return dst;
    }
    //POS_NAME_QUERY_RESPONSE
    static char[] getPosNameQueryResponseTemplate() {
        char dst[] = new char[1405];
        System.arraycopy( COMPOSITE, 0, dst, 0, 1137 );
        System.arraycopy( POS_NAME_QUERY_RESPONSE_TAIL, 0, dst, 1137, 268 );
        return dst;
    }
    //NEG_NAME_QUERY_RESPONSE
    static char[] getNegNameQueryResponseTemplate() {
        char dst[] = new char[1207];
        System.arraycopy( COMPOSITE, 0, dst, 0, 1171 );
        dst[1171] = '\n';
        System.arraycopy( HALF_LINE, 0, dst, 1172, 35 );
        return dst;
    }
    static char[] getRedirectNameQueryResponseTemplate() {
        char dst[] = new char[2078];
        System.arraycopy( COMPOSITE, 0, dst, 0, 2078 );
        return dst;
    }
    static char[] getWackTemplate() {
        char dst[] = new char[1271];
        System.arraycopy( COMPOSITE, 0, dst, 0, 1137 );
        System.arraycopy( WACK_TAIL, 0, dst, 1137, 134 );
        return dst;
    }
    static char[] getNodeStatusResponseTemplate( int numOfNames, int numOfStatBytes ) {
        numOfNames = ( numOfNames > 2 ) ? numOfNames - 2 : 0;
        int statRows = ( numOfStatBytes - 2 ) / 4;
        boolean isHalfEnd = false;
        int size = 1137 + 268 + 67 * numOfNames + 469 + statRows * 134;
        if(( numOfStatBytes - 2 ) % 4 != 0 ) {
            size += 70;
            isHalfEnd = true;
        }
        char dst[] = new char[size];
        System.arraycopy( COMPOSITE, 0, dst, 0, 1137 );
        System.arraycopy( NODE_STATUS_RESPONSE_1, 0, dst, 1137, 268 );
        int offset = 1137 + 268;
        int n = 0;
        int t = ( numOfNames - 1 ) / 2;
        while( n < t ) {
            if( n % 2 == 0 ) {
                System.arraycopy( BAR_LINE, 0, dst, offset, 67 );
                offset += 67;
            } else {
                System.arraycopy( PLUS_LINE, 0, dst, offset, 67 );
                offset += 67;
            }
            n++;
        }
        System.arraycopy( SLASH_LINE, 0, dst, offset, 67 );
        offset += 67;
        if( numOfNames % 2 == 0 ) {
            System.arraycopy( SLASH_LINE, 0, dst, offset, 67 );
            offset += 67;
        }
        n--;
        while( n >= 0 ) {
            if( n % 2 == 0 ) {
                System.arraycopy( BAR_LINE, 0, dst, offset, 67 );
                offset += 67;
            } else {
                System.arraycopy( PLUS_LINE, 0, dst, offset, 67 );
                offset += 67;
            }
            n--;
        }
        System.arraycopy( NODE_STATUS_RESPONSE_2, 0, dst, offset, 469 );
        offset += 469;
        for( int i = 0; i < statRows; i++ ) {
            System.arraycopy( VARIABLE_STATS_1, 0, dst, offset, 134 );
            offset += 134;
        } 
        if( isHalfEnd ) {
            System.arraycopy( VARIABLE_STATS_2, 0, dst, offset, 70 );
            offset += 70;
        }
        return dst;
    }

    static final int P_NAME_TRN_ID = 213;
    static final int P_RESP = 233;
    static final int P_AUTH_ANS = 243;
    static final int P_RCODE = 259;
    static final int P_QUESTI = 347;
    static final int P_ANSWER = 380;
    static final int P_AUTHOR = 481;
    static final int P_ADDITI = 514;
    static final int P_NAME = 672;
    static final int P_TYPE = 880;
    static final int P_CLASS = 911;
    static final int P_TTL = 1036;
    static final int P_RDATA = 1154;
    static final int P_NMFLAGS = 1171;
    static final int P_POS_ADDR = 1298;
    static final int P_NAMES = 1342;
    static final int P_WACK_RESP = 1171;
    static final int P_WACK_AUTH_ANS = 1181;
    static final int P_WACK_RCODE = 1197;
    static final int P_RD_TYPE = 1619;
    static final int P_RD_CLASS = 1649;
    static final int P_RD_TTL = 1773;
    static final int P_RD_RDATA = 1891;
    static final int P_RD_ADDR = 2020;
    
    static void printPacketDiagram( String desc, ClientNameServicePacket nsp ) {
        try {
            if(( PACKET_DIAGRAMS & mask ) == 0 )
                return;

            char buf[];
            Name n = new Name();
            n.readFullName( nsp.bufr, 12 );
            int i, numOfNames = 0, numOfStatBytes = 0, statIndex = 0;
            NbtAddress[] addrs = null;
            int packetType = nsp.getPacketType();
            switch( packetType ) {
                case ClientNameServicePacket.POS_NAME_QUERY_RESPONSE:
                    buf = getPosNameQueryResponseTemplate();
                    break;
                case ClientNameServicePacket.NEG_NAME_QUERY_RESPONSE:
                    buf = getNegNameQueryResponseTemplate();
                    break;
                case ClientNameServicePacket.REDIRECT_NAME_QUERY_RESPONSE:
                    buf = getRedirectNameQueryResponseTemplate();
                    break;
                case ClientNameServicePacket.WAIT_FOR_ACK_RESPONSE:
                    buf = getWackTemplate();
                    break;
                case ClientNameServicePacket.NODE_STATUS_RESPONSE:
                    numOfNames = nsp.bufr[n.wireLength + 22] & 0xFF;
                    addrs = nsp.getNodeNameArray( new NbtAddress( new Name( null, 0, null ), 0, false, 0 ));
                    statIndex = 63 + numOfNames * 18;
                    numOfStatBytes = nsp.getLength() - statIndex; // starts just after MAC (JUMPERS field in RFC)
                    buf = getNodeStatusResponseTemplate( numOfNames, numOfStatBytes );
                    break;
                case ClientNameServicePacket.NAME_QUERY_REQUEST:
                case ClientNameServicePacket.NODE_STATUS_REQUEST:
                    buf = getNameQueryRequestTemplate();
                    break;
                default:
                    return;
            }
            String s = Integer.toString( nsp.getNameTrnId() );
            s.getChars( 0, s.length(), buf, P_NAME_TRN_ID );

            buf[P_RESP] = nsp.isResponse() ? 'R' : '0';

            int opCode = nsp.getOpCode();
            switch( opCode ) {
                case 0:
                    s = new String( "QUERY" );
                    break;
                case 7:
                    s = new String( "WACK" );
                    break;
                default:
                    s = new String( "0x" + getHexString( opCode, 1 ));
            }
            s.getChars( 0, s.length(), buf, P_RESP + 3 );

            buf[P_AUTH_ANS] = nsp.isAuthAnswer() ? 'A' : '0';
            buf[P_AUTH_ANS + 2] = nsp.isTruncated() ? 'T' : ' ';
            buf[P_AUTH_ANS + 4] = nsp.isRecurDesired() ? 'R' : '0';
            buf[P_AUTH_ANS + 6] = nsp.isRecurAvailable() ? 'A' : '0';
            buf[P_AUTH_ANS + 12] = nsp.isBroadcast() ? 'B' : '0';

            i = nsp.getResultCode();
            switch( i ) {
                case 0:
                    s = new String( "0x0" );
                    break;
                case 1:
                    s = new String( "FMT" );
                    break;
                case 2:
                    s = new String( "SRV" );
                    break;
                case 3:
                    s = new String( "NEG" );
                    break;
                case 4:
                    s = new String( "IMP" );
                    break;
                case 5:
                    s = new String( "RFS" );
                    break;
                case 6:
                    s = new String( "ACT" );
                    break;
                case 7:
                    s = new String( "CFT" );
                    break;
                default:
                    s = new String( "0x" + getHexString( i, 1 ));
            }
            s.getChars( 0, s.length(), buf, P_RCODE );

            toHexChars( nsp.getQuestionCount(), buf, P_QUESTI, 4 );
            toHexChars( nsp.getAnswerCount(), buf, P_ANSWER, 4 );
            toHexChars( nsp.getAuthorityCount(), buf, P_AUTHOR, 4 );
            toHexChars( nsp.getAdditionalCount(), buf, P_ADDITI, 4 );

            // name
            if( packetType == ClientNameServicePacket.NODE_STATUS_REQUEST ||
                    packetType == ClientNameServicePacket.NODE_STATUS_RESPONSE ) {
                n.name = new String( "*" );
            }
            s = n.toString();
            s = s.length() > 27 ? s.substring( 0, 27 ) : s;
            s.getChars( 0, s.length(), buf, P_NAME );

            // type
            i = (( nsp.bufr[n.wireLength + 12] << 8 ) & 0xFF ) + ( nsp.bufr[n.wireLength + 13] & 0xFF );
            switch( i ) {
                case 0x01:
                    s = new String( " A (0x0001)" );
                    break;
                case 0x02:
                    s = new String( " NS (0x0002)" );
                    break;
                case 0x0A:
                    s = new String( "NULL (0x000A)" );
                    break;
                case 0x20:
                    s = new String( " NB (0x0020)" );
                    break;
                case 0x21:
                    s = new String( "NBSTAT (0x0021)" );
                    break;
                default:
                    s = new String( "   0x" + getHexString( i, 2 ));
            }
            s.getChars( 0, s.length(), buf, P_TYPE );

            // class
            i = (( nsp.bufr[n.wireLength + 14] << 8 ) & 0xFF ) + ( nsp.bufr[n.wireLength + 15] & 0xFF );
            s = new String( "IN (0x0001)" );
            if( i != 0x1 ) {
                s = new String( "0x" + getHexString( i, 1 ));
            }
            s.getChars( 0, s.length(), buf, P_CLASS );

            // Name Query Request or Node Status Request then we're done
            if( packetType == ClientNameServicePacket.NAME_QUERY_REQUEST || packetType == ClientNameServicePacket.NODE_STATUS_REQUEST ) {
                out.println( desc, buf );
                return;
            }

            // ttl
            s = Integer.toString( nsp.getTTL() );
            i = s.length();
            s.getChars( 0, i, buf, P_TTL - ( i - 1 ) / 2 );
            
            // rdata
            i = (( nsp.bufr[n.wireLength + 20] << 8 ) & 0xFF ) + ( nsp.bufr[n.wireLength + 21] & 0xFF );
            s = Integer.toString( i );
            s.getChars( 0, s.length(), buf, P_RDATA );

            // Neg Name Query Response then we're done
            if( packetType == ClientNameServicePacket.NEG_NAME_QUERY_RESPONSE ) {
                out.println( desc, buf );
                return;
            }

            switch( packetType ) {
                case ClientNameServicePacket.POS_NAME_QUERY_RESPONSE:
                    // name flags
                    buf[P_NMFLAGS] = ( nsp.bufr[n.wireLength + 22] & 0x80 ) == 0 ? 'U' : 'G';
                    switch(( nsp.bufr[n.wireLength + 22] & 0x60 ) >> 5 ) {
                        case 0:
                            buf[P_NMFLAGS + 3] = 'B';
                            break;
                        case 1:
                            buf[P_NMFLAGS + 3] = 'P';
                            break;
                        case 2:
                            buf[P_NMFLAGS + 3] = 'M';
                            break;
                        case 3:
                            buf[P_NMFLAGS + 3] = 'H';
                            break;
                    }
                    // ip name resolved to
                    s = getIP( nsp.bufr, n.wireLength + 24 );
                    s.getChars( 0, s.length(), buf, P_POS_ADDR );
                    break;
                case ClientNameServicePacket.REDIRECT_NAME_QUERY_RESPONSE:
                    // REDIRECT
                    int redirectOffset = nsp.getLength() - 14;
                    // type
                    i = (( nsp.bufr[redirectOffset] << 8 ) & 0xFF ) + ( nsp.bufr[redirectOffset + 1] & 0xFF );
                    switch( i ) {
                        case 0x01:
                            s = new String( " A (0x0001)" );
                            break;
                        case 0x02:
                            s = new String( " NS (0x0002)" );
                            break;
                        case 0x0A:
                            s = new String( "NULL (0x000A)" );
                            break;
                        case 0x20:
                            s = new String( " NB (0x0020)" );
                            break;
                        case 0x21:
                            s = new String( "NBSTAT (0x0021)" );
                            break;
                        default:
                            s = new String( "   0x" + getHexString( i, 2 ));
                    }
                    s.getChars( 0, s.length(), buf, P_RD_TYPE );

                    // class
                    i = (( nsp.bufr[redirectOffset + 2] << 8 ) & 0xFF ) + ( nsp.bufr[redirectOffset + 3] & 0xFF );
                    s = new String( "IN (0x0001)" );
                    if( i != 0x1 ) {
                        s = new String( "0x" + getHexString( i, 2 ));
                    }
                    s.getChars( 0, s.length(), buf, P_RD_CLASS );

                    // ttl
                    i = ( nsp.bufr[redirectOffset + 4] << 24 ) & 0xFF000000;
                    i |= ( nsp.bufr[redirectOffset + 5] << 16 ) & 0xFF0000;
                    i |= ( nsp.bufr[redirectOffset + 6] << 8 ) & 0xFF00;
                    i |= nsp.bufr[redirectOffset + 7] & 0xFF;
                    s = Integer.toString( i );
                    i = s.length();
                    s.getChars( 0, i, buf, P_RD_TTL - ( i - 1 ) / 2 );

                    // rdata
                    i = (( nsp.bufr[redirectOffset + 8] << 8 ) & 0xFF ) + ( nsp.bufr[redirectOffset + 9] & 0xFF );
                    s = Integer.toString( i );
                    s.getChars( 0, s.length(), buf, P_RD_RDATA );

                    // redirect ip
                    s = getIP( nsp.bufr, redirectOffset + 10 );
                    s.getChars( 0, s.length(), buf, P_RD_ADDR );
                    break;
                case ClientNameServicePacket.WAIT_FOR_ACK_RESPONSE:
                    // WACK
                    byte b1 = nsp.bufr[n.wireLength + 22];
                    byte b2 = nsp.bufr[n.wireLength + 23];
                    // response flag
                    buf[P_WACK_RESP] = (( b1 & 0x80 ) == 0x80 ) ? 'R' : '0';

                    // opcode
                    int oC = ( b1 & 0x78 ) >> 3;
                    switch( oC ) {
                        case 0:
                            s = new String( "QUERY" );
                            break;
                        case 7:
                            s = new String( "WACK" );
                            break;
                        default:
                            s = new String( "0x" + getHexString( oC, 1 ));
                    }
                    s.getChars( 0, s.length(), buf, P_WACK_RESP + 3 );

                    // authoritative answer
                    buf[P_WACK_AUTH_ANS    ] = (( b1 & 0x04 ) == 0x04 ) ? 'A' : '0';
                    // truncation
                    buf[P_WACK_AUTH_ANS + 2] = (( b1 & 0x02 ) == 0x02 ) ? 'T' : ' ';
                    // recursion desired
                    buf[P_WACK_AUTH_ANS + 4] = (( b1 & 0x01 ) == 0x01 ) ? 'R' : '0';
                    // recur available
                    buf[P_WACK_AUTH_ANS + 6] = (( b2 & 0x80 ) == 0x80 ) ? 'A' : '0';
                    // broadcast
                    buf[P_WACK_AUTH_ANS +12] = (( b2 & 0x10 ) == 0x10 ) ? 'B' : '0';

                    // result code
                    i = b2 & 0x0F;
                    switch( i ) {
                        case 0:
                            s = new String( "0x0" );
                            break;
                        case 1:
                            s = new String( "FMT" );
                            break;
                        case 2:
                            s = new String( "SRV" );
                            break;
                        case 3:
                            s = new String( "NEG" );
                            break;
                        case 4:
                            s = new String( "IMP" );
                            break;
                        case 5:
                            s = new String( "RFS" );
                            break;
                        case 6:
                            s = new String( "ACT" );
                            break;
                        case 7:
                            s = new String( "CFT" );
                            break;
                        default:
                            s = new String( "0x" + getHexString( i, 1 ));
                    }
                    s.getChars( 0, s.length(), buf, P_WACK_RCODE );
                    break;
                case ClientNameServicePacket.NODE_STATUS_RESPONSE:
                    // NODE_STATUS_RESPONSE
                    // number of names field
                    s = Integer.toString( numOfNames );
                    s.getChars( 0, s.length(), buf, P_RDATA + 24 );
                    // names
                    int pos = P_NAMES;
                    for( int a = 0; a < numOfNames; a++ ) {
                        s = addrs[a].hostName.toString();
                        s.getChars( 0, s.length(), buf, pos );
                        if( addrs[a].groupName ) {
                            new String( "GRP" ).getChars( 0, 3, buf, pos + 32 );
                        } else {
                            new String( "UNQ" ).getChars( 0, 3, buf, pos + 32 );
                        }
                        switch( addrs[a].nodeType ) {
                            case NbtAddress.P_NODE:
                                buf[pos + 36] = 'P';
                                break;
                            case NbtAddress.B_NODE:
                                buf[pos + 36] = 'B';
                                break;
                            case NbtAddress.H_NODE:
                                buf[pos + 36] = 'H';
                                break;
                            case NbtAddress.M_NODE:
                                buf[pos + 36] = 'M';
                                break;
                        }
                        if( addrs[a].isActive ) {
                            new String( "ACT" ).getChars( 0, 3, buf, pos + 38 );
                        } else {
                            new String( "OFF" ).getChars( 0, 3, buf, pos + 38 );
                        }
                        if( addrs[a].isPermanent ) {
                            new String( "PRM" ).getChars( 0, 3, buf, pos + 42 );
                        }
                        if( addrs[a].isInConflict ) {
                            new String( "CNF" ).getChars( 0, 3, buf, pos + 46 );
                        }
                        if( addrs[a].isBeingDeleted ) {
                            new String( "DEL" ).getChars( 0, 3, buf, pos + 50 );
                        }
                        pos += 67;
                    }
                    // mac address
                    byte[] mac = addrs[0].macAddress;
                    pos += 147;
                    for( int m = 0; m < 6; m++ ) {
                        toHexChars( mac[m] & 0xFF, buf, pos, 2 );
                        pos += 3;
                    }
                    // bytes follow
                    pos += 35;
                    s = Integer.toString( numOfStatBytes );
                    s.getChars( 0, s.length(), buf, pos );
                    // bytes total of statistics
                    pos += 67;
                    s = Integer.toString( numOfStatBytes + 6 );
                    s.getChars( 0, s.length(), buf, pos );

                    // stats
                    pos += 44;
                    int b = 0, v = 0;
                    while( b < numOfStatBytes ) {
                        v = nsp.bufr[statIndex    ] & 0xFF;
                        if( v != 0 )
                            toHexChars( v, buf, pos, 2 );
                        v = nsp.bufr[statIndex + 1] & 0xFF;
                        if( v != 0 )
                            toHexChars( v, buf, pos + 2, 2 );
                        statIndex += 2;
                        b += 2;
                        pos = ( b % 4 == 0 ) ? pos + 33 : pos + 101;
                    }
                    break;
            }
            out.println( desc, buf );
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.Log.printPacketDiagram()", e );
        }
    }
    private static String getIP( byte[] src, int srcIndex ) {
        return Integer.toString( src[srcIndex++] & 0xFF ) + "." +
            Integer.toString( src[srcIndex++] & 0xFF ) + "." +
            Integer.toString( src[srcIndex++] & 0xFF ) + "." +
            Integer.toString( src[srcIndex] & 0xFF );
    }
}

