/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.netbios;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Hashtable;

/** 
 * This class represents a netbios over TCP/IP address.
 * 
 * <p> Applications should use the methods <code>getLocalHost</code>,
 * <code>getByName</code>, <code>getAllByName</code>,
 * <code>getAllByAddress</code> to create a new NbtAddress instance. This
 * class is symmetric with {@link java.net.InetAddress}. A netbios
 * host can(and does) have many names associated with it. See {@link
 * jcifs.netbios.NbtNameServiceClient} for details.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see       java.net.InetAddress
 * @since     jcifs-0.1
 */ 

public final class NbtAddress {
    Name hostName = null;
    int address;
    boolean groupName;
    int nodeType;
    boolean isBeingDeleted;
    boolean isInConflict;
    boolean isActive;
    boolean isPermanent;
    byte[] macAddress = null;
    boolean isDataFromNodeStatus = false;

    // save all params for client so that
    // can be intantiated only when there
    // is a request requiring it

    static InetAddress laddr = null;
    static int port = -1;
    static InetAddress nameserver = null;
    static InetAddress bCastAddress = null;
    static int localhostType = -1;
    static String localhostName = null;
    static String localhostScope = null;

    // Node Types

/** 
 * Broadcast queries only. This is the default if a nameserver(WINS) is
 * not provided.
 */ 

    public static final int B_NODE = 0;

/**
 * A Point-to-Point node unicasts queries only. A nameserver must be provided.
 */

    public static final int P_NODE = 1;

/** 
 * Try Broadcast queries first, then try to resolve the name using the
 * nameserver.
 */

    public static final int M_NODE = 2;

/** 
 * A Hybrid node tries to resolve a name using the nameserver first. If that fails use the
 * broadcast address. This is the default if a nameserver is provided. This
 * is the behavior of Windows nodes.
 */ 

    public static final int H_NODE = 3;

    // Usefull Name Strings
    private static final byte[] MASTER_BROWSER_BYTES = {
        (byte)0x01, (byte)0x02, (byte)0x5F, (byte)0x5F, (byte)0x4D, (byte)0x53,
        (byte)0x42, (byte)0x52, (byte)0x4F, (byte)0x57, (byte)0x53, (byte)0x45,
        (byte)0x5F, (byte)0x5F, (byte)0x02
    };
    private static final byte[] ALL_HOSTS_BYTES = {
        (byte)0x2A, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x00
    };

/** 
 * This is a special name for querying the master browser that serves the
 * list of hosts found in "Network Neighborhood".
 */ 

    public static final String MASTER_BROWSER_NAME = new String( MASTER_BROWSER_BYTES );

/** 
 * This is a special name that means all hosts. If you wish to find all hosts
 * on a network querying a workgroup group name is the preferred method.
 */ 

    public static final String ALL_HOSTS_NAME = new String( ALL_HOSTS_BYTES );

    NbtAddress( Name hostName, int address, boolean groupName, int nodeType ) {
        this.hostName = hostName;
        this.address = address;
        this.groupName = groupName;
        this.nodeType = nodeType;
    }
    NbtAddress( Name hostName,
                int address,
                boolean groupName,
                int nodeType,
                boolean isBeingDeleted,
                boolean isInConflict,
                boolean isActive,
                boolean isPermanent,
                byte[] macAddress ) {

        this.hostName = hostName;
        this.address = address;
        this.groupName = groupName;
        this.nodeType = nodeType;
        this.isBeingDeleted = isBeingDeleted;
        this.isInConflict = isInConflict;
        this.isActive = isActive;
        this.isPermanent = isPermanent;
        this.macAddress = macAddress;
        isDataFromNodeStatus = true;
    }

/**
 * Determines if the address is a group address. This is also
 * known as a workgroup name or group name.
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */

    public boolean isGroupAddress() throws UnknownNbtHostException {
        checkData();
        return groupName;
    }

/** 
 * Checks the node type of this address.
 * @return {@link jcifs.netbios.NbtAddress#B_NODE}, {@link jcifs.netbios.NbtAddress#P_NODE}, {@link jcifs.netbios.NbtAddress#M_NODE}, {@link jcifs.netbios.NbtAddress#H_NODE}
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */ 

    public int getNodeType() throws UnknownNbtHostException {
        checkData();
        return nodeType;
    }

/** 
 * Determines if this address in the process of being deleted.
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */ 

    public boolean isBeingDeleted() throws UnknownNbtHostException {
        checkNodeStatusData();
        return isBeingDeleted;
    }

/** 
 * Determines if this address in conflict with another address.
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */ 

    public boolean isInConflict() throws UnknownNbtHostException {
        checkNodeStatusData();
        return isInConflict;
    }


/** 
 * Determines if this address is active.
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */ 


    public boolean isActive() throws UnknownNbtHostException {
        checkNodeStatusData();
        return isActive;
    }


/** 
 * Determines if this address is set to be permanent.
 *
 * @throws UnknownNbtHostException if the host cannot be resolved to find out.
 */ 

    public boolean isPermanent() throws UnknownNbtHostException {
        checkNodeStatusData();
        return isPermanent;
    }


/** 
 * Retrieves the MAC address of the nodes network interface. Samba returns all zeros.
 *
 * @return the MAC address as an array of six bytes
 * @throws UnknownNbtHostException if the host cannot be resolved to determine the MAC address.
 */ 

    public byte[] getMacAddress() throws UnknownNbtHostException {
        checkNodeStatusData();
        return macAddress;
    }

    void checkData() throws UnknownNbtHostException {
        if( hostName.name != null ) {
            return;
        }
        getAllByAddress( this );
    }
    void checkNodeStatusData() throws UnknownNbtHostException {
        if( isDataFromNodeStatus ) {
            return;
        }
        getAllByAddress( this );
    }


/** 
 * The hostname of this address. If the hostname is null the local machines
 * IP address is returned.
 *
 * @return the text representation of the hostname associated with this address
 */ 


    public String getHostName() {
        try {
            if( hostName.name == null ) {
                getAllByAddress( this );
            }
        } catch( UnknownNbtHostException unhe ) {
            return getHostAddress();
        }
        return hostName.toString();
    }


/** 
 * Returns the raw IP address of this NbtAddress. The result is in network
 * byte order: the highest order byte of the address is in getAddress()[0].
 *
 * @return a four byte array
 */ 


    public byte[] getAddress() {    
        byte[] addr = new byte[4];

        addr[0] = (byte)(( address >>> 24 ) & 0xFF );
        addr[1] = (byte)(( address >>> 16 ) & 0xFF );
        addr[2] = (byte)(( address >>> 8 ) & 0xFF );
        addr[3] = (byte)( address & 0xFF );
        return addr;
    }


/** 
 * To convert this address to an <code>InetAddress</code>.
 *
 * @return the {@link java.net.InetAddress} representation of this address.
 */ 


    public InetAddress getInetAddress() throws UnknownHostException {
        return InetAddress.getByName( getHostAddress() );
    }


/** 
 * Returns this IP adress as a {@link java.lang.String} in the form "%d.%d.%d.%d".
 */ 


    public String getHostAddress() {    
        return (( address >>> 24 ) & 0xFF ) + "." +
            (( address >>> 16 ) & 0xFF ) + "." +
            (( address >>> 8 ) & 0xFF ) + "." +
            (( address >>> 0 ) & 0xFF );
    }


/** 
 * Returns a hashcode for this IP address. The hashcode comes from the IP address
 * and is not generated from the string representation. So because netbios nodes
 * can have many names, all names associated with an IP will have the same
 * hashcode.
 */ 


    public int hashCode() {
        return address;
    }


/** 
 * Determines if this address is equal two another. Only the IP Addresses
 * are compared. Similar to the {@link #hashCode} method, the comparison
 * is based on the integer IP address and not the string representation.
 */ 


    public boolean equals( NbtAddress nbAddr ) {
        return ( nbAddr != null ) && ( nbAddr.address == address );
    }


/** 
 * Returns the {@link java.lang.String} representaion of this address.
 */ 

    public String toString() {
        return getHostName() + "/" + getHostAddress();
    }


/** 
 * Retrieves the local host address.
 *
 * @throws UnknownNbtHostException This is not likely as the IP returned by <code>InetAddress</code> should be available
 */ 


    public static NbtAddress getLocalHost() throws UnknownNbtHostException {
        if( localhost == null ) {
            if( client == null ) {
                try {
                    setClient();
                } catch( SocketException se ) {
                    Log.printStackTrace( "jcifs.netbios.NbtAddress exception", se );
                    throw new UnknownNbtHostException( se.getMessage() );
                }
            }
            localhost = client.getLocalHostName();
        }
        return localhost;
    }

    static int cachePolicy = 30;
    static final int FOREVER = -1;
    static Hashtable addressCache = new Hashtable();
    static NbtAddress unknownAddress;
    static NbtAddress[] unknown_array;

    static final class CacheEntry {
        NbtAddress address;
        long expiration;

        CacheEntry( NbtAddress address, long expiration ) {
            this.address = address;
            this.expiration = expiration;
        }
    }

    static void cacheAddress( NbtAddress addr ) {
        if( cachePolicy == 0 ) {
            return;
        }
        long expiration = -1;
        if( cachePolicy != FOREVER ) {
            expiration = System.currentTimeMillis() + cachePolicy * 1000;
        }
        cacheAddress( addr, expiration );
    }
    static void cacheAddress( NbtAddress addr, long expiration ) {
        synchronized( addressCache ) {
            CacheEntry entry = (CacheEntry)addressCache.get( addr.hostName );
            if( entry == null ) {
                entry = new CacheEntry( addr, expiration );
                addressCache.put( addr.hostName, entry );
            } else {
                entry.address = addr;
                entry.expiration = expiration;
            }
            Log.printAddressCache( "name service address cache", addressCache );
        }
    }
    static void cacheAddressArray( NbtAddress[] addrs ) {
        if( cachePolicy == 0 ) {
            return;
        }
        long expiration = -1;
        if( cachePolicy != FOREVER ) {
            expiration = System.currentTimeMillis() + cachePolicy * 1000;
        }
        synchronized( addressCache ) {
            for( int i = 0; i < addrs.length; i++ ) {
                CacheEntry entry = (CacheEntry)addressCache.get( addrs[i].hostName );
                if( entry == null ) {
                    entry = new CacheEntry( addrs[i], expiration );
                    addressCache.put( addrs[i].hostName, entry );
                } else {
                    entry.address = addrs[i];
                    entry.expiration = expiration;
                }
            }
            Log.printAddressCache( "name service address cache", addressCache );
        }
    }
    static NbtAddress getCachedAddress( Name hostName ) {
        synchronized( addressCache ) {
            if( cachePolicy == 0 ) {
                return null;
            }
            CacheEntry entry = (CacheEntry)addressCache.get( hostName );
            if( entry != null && entry.expiration < System.currentTimeMillis() && entry.expiration >= 0 ) {
                entry = null;
            }
            return entry != null ? entry.address : null;
        }
    }

    static NbtNameServiceClient client = null;
    static NbtAddress localhost = null;

    static {
        unknownAddress = new NbtAddress( new Name( "0.0.0.0", 0x00, "" ), 0, false, B_NODE );
        unknown_array = new NbtAddress[1];
        unknown_array[0] = unknownAddress;
        addressCache.put( unknownAddress.hostName, new CacheEntry( unknownAddress, FOREVER ));
    }

    static void setClient() throws SocketException {
        client = new NbtNameServiceClient( port, laddr );
        client.nameserver = nameserver;
        client.bCastAddress = bCastAddress;
        client.nodeType = localhostType;
        client.setLocalHostName( localhostName, localhostScope );
    }

    static NbtAddress doNameQuery( Name name ) throws UnknownNbtHostException {

        NbtAddress addr = getCachedAddress( name );
        if( addr == null ) {
            try {
                if( client == null ) {
                    try {
                        setClient();
                    } catch( SocketException se ) {
                        Log.printStackTrace( "jcifs.netbios.NbtAddress exception", se );
                        throw new UnknownNbtHostException( se.getMessage() );
                    }
                }
                addr = client.getByName( name, true );
            } catch( UnknownNbtHostException unhe ) {
                addr = unknownAddress;
            } finally {
                cacheAddress( addr );
            }
        }
        if( addr == unknownAddress ) {
            throw new UnknownNbtHostException( name.toString() );
        }
        return addr;
    }
    static NbtAddress[] doBroadcastNameQuery( Name name ) throws UnknownNbtHostException {
        NbtAddress addrs[] = null;
        try {
            if( client == null ) {
                try {
                    setClient();
                } catch( SocketException se ) {
                    Log.printStackTrace( "jcifs.netbios.NbtAddress exception", se );
                    throw new UnknownNbtHostException( se.getMessage() );
                }
            }
            addrs = client.getAllByName( name, true );
        } catch( UnknownNbtHostException unhe ) {
            addrs = unknown_array;
            throw new UnknownNbtHostException( name.toString() );
        } finally {
            cacheAddressArray( addrs );
        }
        return addrs;
    }

/** 
 * Determines the address of a host given it's host name. The name can be a
 * machine name like "jcifs.samba.org" or an IP address like "192.168.1.15".
 *
 * @param host hostname to resolve
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public static NbtAddress getByName( String host )
                                        throws UnknownNbtHostException {
        return getByName( host, 0, null );
    }

/** 
 * Determines the address of a host given it's host name. Netbios
 * names also have a <code>type</code>. Types(aka Hex Codes)
 * are used to distiquish the various services on a host. <a
 * href="http://users.erols.com/mballen/NetBiosHexCodes.txt">Here</a> is
 * a fairly complete list of netbios hex codes. Scope is not used but is
 * still functional in other netbios products and so for completeness it has been
 * implemented. A <code>scope</code> of <code>null</code> or <code>""</code>
 * signifies no scope.
 *
 * @param host the name to resolve
 * @param type the hex code of the name
 * @param scope the scope of the name
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public static NbtAddress getByName( String host,
                                        int type,
                                        String scope )
                                        throws UnknownNbtHostException {

        if( host == null || host.length() == 0 ) {
            return getLocalHost();
        }
        if( !Character.isDigit( host.charAt(0) )) {
            return (NbtAddress)doNameQuery( new Name( host, type, scope ));
        } else {
            int IP = 0x00;
            int hitDots = 0;
            char[] data = host.toCharArray();

            for( int i = 0; i < data.length; i++ ) {
                char c = data[i];
                if( c < 48 || c > 57 ) {
                    return (NbtAddress)doNameQuery( new Name( host, type, scope ));
                }
                int b = 0x00;
                while( c != '.' ) {
                    if( c < 48 || c > 57 ) {
                        return (NbtAddress)doNameQuery( new Name( host, type, scope ));
                     }
                     b = b * 10 + c - '0';

                    if( ++i >= data.length )
                        break;

                    c = data[i];
                }
                if( b > 0xFF ) {
                    return (NbtAddress)doNameQuery( new Name( host, type, scope ));
                }
                IP = ( IP << 8 ) + b;
                hitDots++;
            }
            if( hitDots != 4 || host.endsWith( "." )) {
                return (NbtAddress)doNameQuery( new Name( host, type, scope ));
            }
            return new NbtAddress( new Name( null, type, scope ), IP, false, B_NODE );
        }
    }

/** 
 * Resloves the addresses of all hosts given a name(presumably a workgroup
 * name).
 *
 * @param host groupname to lookup
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public static NbtAddress[] getAllByName( String host )
                                        throws UnknownNbtHostException {
        return getAllByName( host, 0, null );
    }

/** 
 * Resloves the addresses of all hosts given a hostname(presumably a
 * workgroup name).  See {@link #getByName}
 * for a description of <code>type</code> and <code>scope</code>.
 *
 * @param host group name to lookup
 * @param type the hexcode of the name
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public static NbtAddress[] getAllByName( String host,
                                        int type,
                                        String scope )
                                        throws UnknownNbtHostException {

        if( host == null || host.length() == 0 ) {
            throw new UnknownNbtHostException( "empty string" );
        }

        if( Character.isDigit( host.charAt( 0 ))) {
            NbtAddress[] ret = new NbtAddress[1];
            ret[0] = getByName( host, type, scope );
            return ret;
        } else {
            NbtAddress[] addrs = doBroadcastNameQuery( new Name( host, type, scope ));
            if( addrs == unknown_array ) {
                throw new UnknownNbtHostException();
            }
            return addrs;
        }
    }

/**
 * Retrieve all addresses of a host by it's address. Netbios hosts can
 * have many names for a given IP address. The name and IP address make the
 * netbios address. This provides a way to retrieve the other names for a
 * host with the same IP address.
 *
 * @param host hostname to lookup all addresses for
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */


    public static NbtAddress[] getAllByAddress( String host ) throws UnknownNbtHostException {
        return getAllByAddress( getByName( host, 0x00, "" ));
    }


/**
 * Retrieve all addresses of a host by it's address. Netbios hosts can
 * have many names for a given IP address. The name and IP address make
 * the netbios address. This provides a way to retrieve the other names
 * for a host with the same IP address.  See {@link #getByName}
 * for a description of <code>type</code>
 * and <code>scope</code>.
 *
 * @param host hostname to lookup all addresses for
 * @param type the hexcode of the name
 * @param scope the scope of the name
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */


    public static NbtAddress[] getAllByAddress( String host,
                                        int type,
                                        String scope )
                                        throws UnknownNbtHostException {
        return getAllByAddress( getByName( host, type, scope ));
    }


/**
 * Retrieve all addresses of a host by it's address. Netbios hosts can
 * have many names for a given IP address. The name and IP address make the
 * netbios address. This provides a way to retrieve the other names for a
 * host with the same IP address.
 *
 * @param addr the address to query
 * @throws UnknownNbtHostException if address cannot be resolved
 */

    public static NbtAddress[] getAllByAddress( NbtAddress addr ) throws UnknownNbtHostException {
        try {
            if( client == null ) {
                try {
                    setClient();
                } catch( SocketException se ) {
                    Log.printStackTrace( "jcifs.netbios.NbtAddress exception", se );
                    throw new UnknownNbtHostException( se.getMessage() );
                }
            }
            NbtAddress[] addrs = client.getNodeStatus( addr );
            cacheAddressArray( addrs );
            return addrs;
        } catch( UnknownNbtHostException unhe ) {
            throw new UnknownNbtHostException( "no name with type 0x" +
                                    Name.toHexChars( addr.hostName.type ) +
                                    ( addr.hostName.scope != null ? " with scope " + addr.hostName.scope : " with no scope" ) +
                                    " for host " + addr.getHostAddress() );
        }
    }
}

